package com.ningdatech.pmapi.filemanage.manage;

import cn.hutool.core.collection.CollUtil;
import com.baomidou.mybatisplus.core.toolkit.Wrappers;
import com.baomidou.mybatisplus.extension.plugins.pagination.Page;
import com.google.common.collect.Lists;
import com.ningdatech.basic.function.VUtils;
import com.ningdatech.basic.model.PageVo;
import com.ningdatech.pmapi.common.constant.BizConst;
import com.ningdatech.pmapi.common.enumeration.ProjectProcessStageEnum;
import com.ningdatech.pmapi.common.helper.UserInfoHelper;
import com.ningdatech.pmapi.filemanage.model.entity.ProjectApplyBorrow;
import com.ningdatech.pmapi.filemanage.model.param.ProjectFileListParam;
import com.ningdatech.pmapi.filemanage.model.vo.ProjectFileListVO;
import com.ningdatech.pmapi.filemanage.model.vo.ProjectFileVO;
import com.ningdatech.pmapi.filemanage.service.INdProjectApplyBorrowService;
import com.ningdatech.pmapi.organization.model.entity.DingOrganization;
import com.ningdatech.pmapi.organization.service.IDingOrganizationService;
import com.ningdatech.pmapi.projectdeclared.manage.DefaultDeclaredProjectManage;
import com.ningdatech.pmapi.projectlib.enumeration.InstTypeEnum;
import com.ningdatech.pmapi.projectlib.enumeration.ProjectStatusEnum;
import com.ningdatech.pmapi.projectlib.manage.ProjectLibManage;
import com.ningdatech.pmapi.projectlib.model.entity.Project;
import com.ningdatech.pmapi.projectlib.model.entity.ProjectInst;
import com.ningdatech.pmapi.projectlib.model.vo.ProjectDetailVO;
import com.ningdatech.pmapi.projectlib.service.IProjectInstService;
import com.ningdatech.pmapi.projectlib.service.IProjectService;
import com.ningdatech.pmapi.staging.enums.MsgTypeEnum;
import com.ningdatech.pmapi.sys.manage.NoticeManage;
import com.ningdatech.pmapi.todocenter.constant.WorkNoticeContant;
import com.ningdatech.pmapi.user.security.auth.model.UserFullInfoDTO;
import com.ningdatech.pmapi.user.util.LoginUserUtil;
import com.wflow.bean.entity.WflowModels;
import com.wflow.contants.HisProInsEndActId;
import com.wflow.exception.BusinessException;
import com.wflow.workflow.bean.dto.OrgInfoDTO;
import com.wflow.workflow.bean.vo.ProcessStartParamsVo;
import com.wflow.workflow.service.ProcessInstanceService;
import com.wflow.workflow.service.ProcessModelService;
import lombok.AllArgsConstructor;
import lombok.extern.slf4j.Slf4j;
import org.apache.commons.lang3.StringUtils;
import org.checkerframework.checker.nullness.qual.Nullable;
import org.flowable.engine.HistoryService;
import org.flowable.engine.history.HistoricProcessInstance;
import org.springframework.stereotype.Component;

import java.time.LocalDateTime;
import java.util.*;
import java.util.stream.Collectors;

/**
 * @Classname ProjectFileManage
 * @Description
 * @Date 2023/6/6 16:04
 * @Author PoffyZhang
 */
@Component
@AllArgsConstructor
@Slf4j
public class ProjectFileManage {

    private final IProjectService projectService;
    private final ProjectLibManage projectLibManage;
    private final UserInfoHelper userInfoHelper;
    private final ProcessModelService processModelService;
    private final DefaultDeclaredProjectManage defaultDeclaredProjectManage;
    private final ProcessInstanceService processInstanceService;
    private final NoticeManage noticeManage;
    private final IProjectInstService projectInstService;
    private final HistoryService historyService;
    private final INdProjectApplyBorrowService projectApplyBorrowService;
    private final IDingOrganizationService dingOrganizationService;

    public PageVo<ProjectFileListVO> list(ProjectFileListParam param) {

        UserFullInfoDTO userFullInfo = userInfoHelper.getUserFullInfo(LoginUserUtil.getUserId());
        String employeeCode = userFullInfo.getEmployeeCode();
        String empPosUnitCode = userFullInfo.getEmpPosUnitCode();
        Integer pageNumber = param.getPageNumber();
        Integer pageSize = param.getPageSize();

        Set<String> subOrgSet = getSubOrgList(empPosUnitCode);
        List<Project> projectList = projectService.list(Wrappers.lambdaQuery(Project.class)
            .like(StringUtils.isNotBlank(param.getProjectName()), Project::getProjectName, param.getProjectName())
            .like(StringUtils.isNotBlank(param.getBuildOrgName()), Project::getBuildOrgName, param.getBuildOrgName())
            .eq(Project::getStatus, ProjectStatusEnum.ACCEPTED.getCode()).orderByDesc(Project::getUpdateOn));

        // 从申请借阅信息表中查出本单位及下属单位审批通过的项目添加到列表中
        HashSet<Long> borrowProjectIdSet = new HashSet<>();
        List<ProjectApplyBorrow> applyBorrowList = projectApplyBorrowService
            .list(Wrappers.lambdaQuery(ProjectApplyBorrow.class)
                .in(ProjectApplyBorrow::getApplyBorrowEmpPosUnitCode, subOrgSet)
                .eq(ProjectApplyBorrow::getIsSuccess, Boolean.TRUE))
            .stream().filter(a -> borrowProjectIdSet.add(a.getProjectId())).collect(Collectors.toList());
        for (ProjectApplyBorrow applyBorrow : applyBorrowList) {
            Long projectId = applyBorrow.getProjectId();
            if (Objects.nonNull(projectId)) {
                Project project = projectService.getById(projectId);
                projectList.add(project);
            }
        }
        // 对整合后的档案列表去重
        HashSet<Long> projectIdSet = new HashSet<>();
        List<Project> projects =
            projectList.stream().filter(p -> projectIdSet.add(p.getId())).collect(Collectors.toList());

        if (CollUtil.isEmpty(projects)) {
            return PageVo.empty();
        }

        List<Project> records =
            projects.stream().skip((long)(pageNumber - 1) * pageSize).limit(pageSize).collect(Collectors.toList());

        List<ProjectFileListVO> vos = records.stream().map(p -> {
            ProjectFileListVO vo = new ProjectFileListVO();
            vo.setId(p.getId());
            // 用户只可查阅自己及下属单位的项目档案；
            if (subOrgSet.contains(p.getBuildOrgCode())) {
                vo.setCanRead(Boolean.TRUE);
            }
            vo.setBuildOrgName(p.getBuildOrgName());
            vo.setPojectCode(p.getProjectCode());
            vo.setProjectName(p.getProjectName());
            // 判断当前项目,当前登录用户是否已经开启了申请借阅并且没有审批完成
            String instCode = p.getInstCode();
            // 获取流程实例
            HistoricProcessInstance newInstance =
                historyService.createHistoricProcessInstanceQuery().processInstanceId(instCode).singleResult();
            ProjectInst projectInst = projectInstService
                .getOne(Wrappers.lambdaQuery(ProjectInst.class).eq(ProjectInst::getInstCode, instCode));
            Integer instType = projectInst.getInstType();
            // 从申请借阅信息表中查询当前项目是否有当前登录用户发起的申请借阅审批流程
            ProjectApplyBorrow applyBorrow = projectApplyBorrowService.getOne(Wrappers
                .lambdaQuery(ProjectApplyBorrow.class).eq(ProjectApplyBorrow::getApplyBorrowEmployeeCode, employeeCode)
                .eq(ProjectApplyBorrow::getProjectId, p.getId()).eq(ProjectApplyBorrow::getInstanceId, instCode));
            if (Objects.nonNull(applyBorrow) && InstTypeEnum.APPLY_BORROW.getCode().equals(instType)
                && !HisProInsEndActId.END.equals(newInstance.getEndActivityId())) {
                vo.setCanApplyBorrow(Boolean.FALSE);
            }
            return vo;
        }).collect(Collectors.toList());

        return PageVo.of(vos, projectList.size());
    }

    private Set<String> getSubOrgList(String empPosUnitCode) {
        HashSet<String> orgSet = new HashSet<>();
        // 先将自己加入子集
        orgSet.add(empPosUnitCode);
        Set<String> subOrgList = dingOrganizationService
            .list(Wrappers.lambdaQuery(DingOrganization.class).eq(DingOrganization::getParentCode, empPosUnitCode))
            .stream().map(DingOrganization::getOrganizationCode).collect(Collectors.toSet());
        // 遍历子集
        for (String orgCode : subOrgList) {
            orgSet.addAll(getSubOrgList(orgCode));
        }
        return orgSet;
    }

    public ProjectFileVO file(Long projectId) {
        ProjectDetailVO projectDetailVo = projectLibManage.getProjectDetail(projectId);

        return null;
    }

    public String wantRead(Long projectId) {
        return "申请成功";
    }

    public String startApplyBorrowProcess(Long projectId) {
        UserFullInfoDTO user = userInfoHelper.getUserFullInfo(LoginUserUtil.getUserId());
        String employeeCode = user.getEmployeeCode();
        VUtils.isTrue(StringUtils.isBlank(employeeCode)).throwMessage("获取登录用户 员工号 失败!");

        Project project = projectService.getNewProject(projectId);
        VUtils.isTrue(Objects.isNull(project)).throwMessage("提交失败 此项目不存在!");

        // 首先要判断 项目当前状态 是不是 已归档
        VUtils.isTrue(!ProjectStatusEnum.ACCEPTED.getCode().equals(project.getStatus())).throwMessage("提交失败 该项目不是 已归档");

        String regionCode = user.getRegionCode();

        WflowModels model = processModelService.getOne(Wrappers.lambdaQuery(WflowModels.class)
            .eq(WflowModels::getRegionCode, regionCode)
            .eq(WflowModels::getProcessType, ProjectProcessStageEnum.APPLY_BORROW.getCode()).last(BizConst.LIMIT_1));

        if (Objects.isNull(model)) {
            log.error("此 【{}】区域找不到申请借阅流程配置", regionCode);
            throw new BusinessException(String.format("此 【%s】区域找不到申请借阅流程配置", regionCode));
        }

        ProcessStartParamsVo params = new ProcessStartParamsVo();
        params.setUser(defaultDeclaredProjectManage.buildUser(employeeCode));
        params.setProcessUsers(Collections.emptyMap());
        // 放入条件判断的项目字段
        // 把条件值给放入工作流
        defaultDeclaredProjectManage.buildCondition(params, project);
        // 获取发起单位、发起单位主管单位、发起单位上级条线主管单位信息
        Map<String, OrgInfoDTO> orgModelMap = defaultDeclaredProjectManage.buildOrgModelMap(employeeCode, project);
        String instanceId =
            processInstanceService.newStartProcess(model.getProcessDefId(), model.getFormId(), params, orgModelMap);
        log.info("申请借阅申报成功 【{}】", instanceId);

        // 保存项目借阅实例相关
        projectLibManage.saveProjectByApplyBorrow(project, user, instanceId);

        // 发送给第一个审批人消息
        noticeManage.sendFirtUser(project, model.getFormName(), instanceId, WorkNoticeContant.PASS_MSG_TEMPLATE,
            MsgTypeEnum.PROJECT_REVIEW);

        return instanceId;
    }
}
